/**
 * OW2 Util
 * Copyright (C) 2008-2011 Bull S.A.S.
 * Contact: easybeans@ow2.org
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
* USA
*
* --------------------------------------------------------------------------
* $Id: OSGiDeployer.java 5932 2011-08-04 11:47:28Z benoitf $
* --------------------------------------------------------------------------
*/

package org.ow2.util.ee.deploy.impl.osgi;

import java.util.HashMap;
import java.util.Map;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleException;
import org.osgi.framework.ServiceReference;
import org.osgi.service.startlevel.StartLevel;
import org.ow2.util.ee.deploy.api.deployable.IDeployable;
import org.ow2.util.ee.deploy.api.deployable.OSGiDeployable;
import org.ow2.util.ee.deploy.api.deployer.DeployerException;
import org.ow2.util.ee.deploy.impl.deployer.AbsDeployer;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * OSGi deployer that is able to deploy an OSGi Deployable
 * @author Florent BENOIT
 */
public class OSGiDeployer extends AbsDeployer {

    /**
     * Logger.
     */
    private Log logger = LogFactory.getLog(OSGiDeployer.class);

    /**
     * Bundle Context to use for the deployment.
     */
    private BundleContext bundleContext = null;

    /**
     * List of Bundles that have been installed.
     */
    private Map<String, Bundle> installedBundles = null;

    /**
     * OSGi StartLevel service reference.
     */
    private StartLevel startLevel = null;

    /**
     * Default constructor.
     * @param bundleContext OSGi bundle context
     */
    public OSGiDeployer(final BundleContext bundleContext) {
        this.bundleContext = bundleContext;
        this.installedBundles = new HashMap<String, Bundle>();

        // Get the StartLevel service reference
        ServiceReference serviceReference = bundleContext.getServiceReference(StartLevel.class.getName());
        setStartLevel((StartLevel) bundleContext.getService(serviceReference));
    }

    /**
     * @param service the StartLevel service
     */
    public void setStartLevel(StartLevel service) {
        this.startLevel = service;
    }

    /**
     * Undeploy the given Bundle.
     * @param osgiDeployable the deployable to remove.
     * @throws DeployerException if the bundle is not undeployed.
     */
    public void doUndeploy(final IDeployable<?> osgiDeployable) throws DeployerException {
        logger.debug("Undeploying {0}", osgiDeployable);

        // Bundle that will be handled
        Bundle osgiBundle = null;

        // Check if it has been installed ?
        if (!isDeployed(osgiDeployable)) {
            throw new DeployerException("Cannot undeploy the bundle '" + osgiDeployable + "' as it was not deployed.");
        }

        // get the bundle Object
        osgiBundle = installedBundles.get(getURL(osgiDeployable).toExternalForm());

        // Check found
        if (osgiBundle == null) {
            throw new DeployerException("No bundle found for the '" + osgiDeployable + "'.");
        }

        // Do not stop or uninstall if already uninstall/removed
        if (Bundle.UNINSTALLED != osgiBundle.getState()) {

            // Stop the bundle
            try {
                osgiBundle.stop();
            } catch (BundleException e) {
                throw new DeployerException("Cannot stop the bundle '" + osgiDeployable + "'.", e);
            }

            // Uninstall the bundle
            try {
                osgiBundle.uninstall();
            } catch (BundleException e) {
                throw new DeployerException("Cannot uninstall the bundle '" + osgiDeployable + "'.", e);
            }
        }

        // Remove it of installed bundle
        installedBundles.remove(getURL(osgiDeployable).toExternalForm());
    }


    /**
     * Deploy the given Bundle.
     * @param deployable the deployable to add.
     * @throws DeployerException if the Bundle is not deployed.
     */
    public void doDeploy(final IDeployable<?> deployable) throws DeployerException {
        logger.debug("Deploying {0}", deployable);

        // Bundle that will be handled
        Bundle osgiBundle = null;

        // Cast the deployable only once
        OSGiDeployable osgiDeployable = OSGiDeployable.class.cast(deployable);

        // Set the initial bundle start level
        startLevel.setInitialBundleStartLevel(osgiDeployable.getStartLevel());

        // First, install the bundle
        try {
            String url = getURL(deployable).toExternalForm();
            if (osgiDeployable.isReference()) {
                url = "reference:".concat(url);
            }
            osgiBundle = bundleContext.installBundle(url);
        } catch (BundleException e) {
            // Check the bundle exception type
            if (e.getType() == BundleException.DUPLICATE_BUNDLE_ERROR) {
                logger.warn("Cannot deploy a bundle that is already installed", e.getMessage());
                return;
            } else {
                throw new DeployerException("Cannot install the bundle '" + osgiDeployable + "'.", e);
            }
        }

        if (osgiDeployable.isStart()) {
            // And then start it
            try {
                if (osgiDeployable.isStartTransient()) {
                    osgiBundle.start(Bundle.START_TRANSIENT);
                } else {
                    osgiBundle.start();
                }
            } catch (BundleException e) {
                logger.error("Cannot start the bundle ''{0}''.", deployable, e);
            }
        }

        // Add it as installed
        installedBundles.put(getURL(deployable).toExternalForm(), osgiBundle);
    }

    /**
     * Checks if the given deployable is supported by the Deployer.
     * @param deployable the deployable to be checked
     * @return true if it is supported, else false.
     */
    public boolean supports(final IDeployable<?> deployable) {
        return OSGiDeployable.class.isInstance(deployable);
    }

    /**
     * Check if the given deployable is deployable or not.
     * @param deployable the deployable to check.
     * @throws DeployerException if the deployable is not supported.
     */
    protected void check(final IDeployable<?> deployable) throws DeployerException {
        if (bundleContext == null) {
            throw new DeployerException("The deployment of the deployable'" + deployable
                    + "' is not available as there is no bundle context set on this deployer.");
        }

        super.check(deployable);
    }
}
